/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __SRC_TABRMD_GENERATED_H__
#define __SRC_TABRMD_GENERATED_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for com.intel.tss2.TctiTabrmd */

#define TYPE_TCTI_TABRMD (tcti_tabrmd_get_type ())
#define TCTI_TABRMD(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_TCTI_TABRMD, TctiTabrmd))
#define IS_TCTI_TABRMD(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_TCTI_TABRMD))
#define TCTI_TABRMD_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), TYPE_TCTI_TABRMD, TctiTabrmdIface))

struct _TctiTabrmd;
typedef struct _TctiTabrmd TctiTabrmd;
typedef struct _TctiTabrmdIface TctiTabrmdIface;

struct _TctiTabrmdIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_cancel) (
    TctiTabrmd *object,
    GDBusMethodInvocation *invocation,
    guint64 arg_id);

  gboolean (*handle_create_connection) (
    TctiTabrmd *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_set_locality) (
    TctiTabrmd *object,
    GDBusMethodInvocation *invocation,
    guint64 arg_id,
    guchar arg_locality);

};

GType tcti_tabrmd_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *tcti_tabrmd_interface_info (void);
guint tcti_tabrmd_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void tcti_tabrmd_complete_create_connection (
    TctiTabrmd *object,
    GDBusMethodInvocation *invocation,
    guint64 id);

void tcti_tabrmd_complete_cancel (
    TctiTabrmd *object,
    GDBusMethodInvocation *invocation,
    guint return_code);

void tcti_tabrmd_complete_set_locality (
    TctiTabrmd *object,
    GDBusMethodInvocation *invocation,
    guint return_code);



/* D-Bus method calls: */
void tcti_tabrmd_call_create_connection (
    TctiTabrmd *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean tcti_tabrmd_call_create_connection_finish (
    TctiTabrmd *proxy,
    guint64 *out_id,
    GAsyncResult *res,
    GError **error);

gboolean tcti_tabrmd_call_create_connection_sync (
    TctiTabrmd *proxy,
    guint64 *out_id,
    GCancellable *cancellable,
    GError **error);

void tcti_tabrmd_call_cancel (
    TctiTabrmd *proxy,
    guint64 arg_id,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean tcti_tabrmd_call_cancel_finish (
    TctiTabrmd *proxy,
    guint *out_return_code,
    GAsyncResult *res,
    GError **error);

gboolean tcti_tabrmd_call_cancel_sync (
    TctiTabrmd *proxy,
    guint64 arg_id,
    guint *out_return_code,
    GCancellable *cancellable,
    GError **error);

void tcti_tabrmd_call_set_locality (
    TctiTabrmd *proxy,
    guint64 arg_id,
    guchar arg_locality,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean tcti_tabrmd_call_set_locality_finish (
    TctiTabrmd *proxy,
    guint *out_return_code,
    GAsyncResult *res,
    GError **error);

gboolean tcti_tabrmd_call_set_locality_sync (
    TctiTabrmd *proxy,
    guint64 arg_id,
    guchar arg_locality,
    guint *out_return_code,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define TYPE_TCTI_TABRMD_PROXY (tcti_tabrmd_proxy_get_type ())
#define TCTI_TABRMD_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_TCTI_TABRMD_PROXY, TctiTabrmdProxy))
#define TCTI_TABRMD_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_TCTI_TABRMD_PROXY, TctiTabrmdProxyClass))
#define TCTI_TABRMD_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_TCTI_TABRMD_PROXY, TctiTabrmdProxyClass))
#define IS_TCTI_TABRMD_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_TCTI_TABRMD_PROXY))
#define IS_TCTI_TABRMD_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_TCTI_TABRMD_PROXY))

typedef struct _TctiTabrmdProxy TctiTabrmdProxy;
typedef struct _TctiTabrmdProxyClass TctiTabrmdProxyClass;
typedef struct _TctiTabrmdProxyPrivate TctiTabrmdProxyPrivate;

struct _TctiTabrmdProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  TctiTabrmdProxyPrivate *priv;
};

struct _TctiTabrmdProxyClass
{
  GDBusProxyClass parent_class;
};

GType tcti_tabrmd_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (TctiTabrmdProxy, g_object_unref)
#endif

void tcti_tabrmd_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
TctiTabrmd *tcti_tabrmd_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
TctiTabrmd *tcti_tabrmd_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void tcti_tabrmd_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
TctiTabrmd *tcti_tabrmd_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
TctiTabrmd *tcti_tabrmd_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define TYPE_TCTI_TABRMD_SKELETON (tcti_tabrmd_skeleton_get_type ())
#define TCTI_TABRMD_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_TCTI_TABRMD_SKELETON, TctiTabrmdSkeleton))
#define TCTI_TABRMD_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_TCTI_TABRMD_SKELETON, TctiTabrmdSkeletonClass))
#define TCTI_TABRMD_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_TCTI_TABRMD_SKELETON, TctiTabrmdSkeletonClass))
#define IS_TCTI_TABRMD_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_TCTI_TABRMD_SKELETON))
#define IS_TCTI_TABRMD_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_TCTI_TABRMD_SKELETON))

typedef struct _TctiTabrmdSkeleton TctiTabrmdSkeleton;
typedef struct _TctiTabrmdSkeletonClass TctiTabrmdSkeletonClass;
typedef struct _TctiTabrmdSkeletonPrivate TctiTabrmdSkeletonPrivate;

struct _TctiTabrmdSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  TctiTabrmdSkeletonPrivate *priv;
};

struct _TctiTabrmdSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType tcti_tabrmd_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (TctiTabrmdSkeleton, g_object_unref)
#endif

TctiTabrmd *tcti_tabrmd_skeleton_new (void);


G_END_DECLS

#endif /* __SRC_TABRMD_GENERATED_H__ */
